/*

Stats-Collector schema definition for MySql.


PUBLIC TABLES (shared read access)

fasp_transfers  stores information about a transfer. A transfer is either a
                single session or multiple sessions with retry.
fasp_sessions   stores information about a session.
fasp_stats      stores statistics (byte counts, etc.) for both sessions and
                transfers. The stats for transfers are the aggregated values of
                all sessions of that transfer. If a transfer consists of a
                single session, both transfer and session point to the same
                stats entry.
fasp_files      stores file info and statistics. Only one entry is stored even
                if a file is transferred in multiple sessions (within the same
                transfer). The stats represent the values of the last session
                that transferred (parts) of the file.
                => The current schema does not provide aggregate file stats.
fasp_nodes      information about the nodes polled by stats-collector.
api_commands    API interface to add/modify/remove nodes.
configuration   configuration properties (other than nodes).

RELATIONSHIPS

fasp_transfers  1-->n fasp_sessions
                1-->n fasp_files
                1-->1 fasp_stats
fasp_sessions   1-->n fasp_files
                1-->1 fasp_stats


INTERNAL TABLES

sc_iterator_sequence    iterator sequence for seq_created/seq_modified
sc_version              schema version information
sc_nodes_poll_config    configuration information for accessing node API on the
                        nodes that are polled.
sc_nodes_poll_status    polling status for each node.

RELATIONSHIPS

sc_nodes_poll_config    1-->n fasp_nodes (poll config uses DNS name, fasp_nodes
                                          contains 1 entry per resolved IP
                                          address)

*/


SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT;
SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS;
SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION;
SET NAMES utf8;
SET @OLD_TIME_ZONE=@@TIME_ZONE;
SET TIME_ZONE='+00:00';
SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0;
SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0;
SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO';
SET @OLD_SQL_NOTES=@@SQL_NOTES, SQL_NOTES=0;

USE /*DBNAME*/`aspera_stats_collector`/*DBNAME*/;

CREATE TABLE `fasp_files` (
  `id`                  bigint          NOT NULL auto_increment,
  `seq_created`         bigint,
  `seq_modified`        bigint,
  `file_id`             varbinary(127)  NOT NULL,
  `transfer_id`         varbinary(127)  NOT NULL,
  `session_id`          varbinary(127)  NOT NULL,               /* the last session that actually transferred (parts of) the file or first session that reported the file as skipped */

  `session_count`       int             NOT NULL DEFAULT 0,     /* the number of sessions that contributed to the transfer of this file. 0 if skipped in all sessions. */
  `status`              enum('RUNNING','COMPLETED','ERROR','UNKNOWN') NOT NULL,
  `started_at`          datetime        NOT NULL,
  `stopped_at`          datetime,
  `updated_at`          datetime,

  `client_path`         text,                                   /* full file path. */
  `server_path`         text,                                   /* full file path. */
  `size`                bigint          NOT NULL DEFAULT 0,

  `start_byte`          bigint          NOT NULL DEFAULT 0,
  `bytes_written`       bigint          NOT NULL DEFAULT 0,
  `bytes_contig`        bigint          NOT NULL DEFAULT 0,
  `bytes_lost`          bigint          NOT NULL DEFAULT 0,

  `elapsed_usec`        bigint          NOT NULL DEFAULT 0,
  `err_code`            int             NOT NULL DEFAULT 0,
  `err_desc`            varchar(255),
  `checksum`            varbinary(255),
  `checksum_type`       enum('NONE','SHA1','MD5','UNKNOWN') DEFAULT 'NONE',

  PRIMARY KEY  (`id`),
  UNIQUE KEY `idx_unique` (`transfer_id`,`file_id`),
  UNIQUE KEY `idx_seq_created` (`seq_created`),
  UNIQUE KEY `idx_seq_modified` (`seq_modified`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

CREATE TABLE `fasp_sessions` (
  `id`                  bigint          NOT NULL auto_increment,
  `session_id`          varbinary(127)  NOT NULL,
  `stats_id`            bigint          NOT NULL,               /* foreign key to fasp_stats.id */
  `transfer_id`         varbinary(127)  NOT NULL,

  `server_node_id`      varbinary(127),
  `server_ip`           varchar(255),
  `server_ssh_port`     int             NOT NULL DEFAULT 0,
  `server_fasp_port`    int             NOT NULL DEFAULT 0,
  `client_node_id`      varbinary(127),
  `client_ip`           varchar(255),
  `client_fasp_port`    int             NOT NULL DEFAULT 0,

  `args_attempted`      int             NOT NULL DEFAULT 0,
  `args_completed`      int             NOT NULL DEFAULT 0,
  `paths_attempted`     int             NOT NULL DEFAULT 0,
  `paths_failed`        int             NOT NULL DEFAULT 0,
  `paths_irreg`         int             NOT NULL DEFAULT 0,
  `paths_excluded`      int             NOT NULL DEFAULT 0,
  `dirscans_completed`  int             NOT NULL DEFAULT 0,
  `filescans_completed` int             NOT NULL DEFAULT 0,
  `mkdirs_attempted`    int             NOT NULL DEFAULT 0,
  `mkdirs_failed`       int             NOT NULL DEFAULT 0,
  `mkdirs_passed`       int             NOT NULL DEFAULT 0,

  PRIMARY KEY  (`id`),
  UNIQUE KEY `idx_session_id` (`session_id`),
  UNIQUE KEY `idx_stats_id` (`stats_id`),
  KEY `idx_transfer_id` (`transfer_id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

CREATE TABLE `fasp_transfers` (
  `id`                  bigint          NOT NULL auto_increment,
  `transfer_id`         varbinary(127)  NOT NULL,
  `stats_id`            bigint          NOT NULL,               /* foreign key to fasp_stats.id */

  `reported_by`         enum('CLIENT','SERVER','CLIENT_AND_SERVER','UNKNOWN') NOT NULL,
  `last_session_id`     varbinary(127)  NOT NULL,
  `session_count`       int             NOT NULL,
  `retry_timeout`       bigint          NOT NULL,               /* if timeout exceeded, client will not try again */
  `client_cluster_id`   varbinary(127),
  `server_cluster_id`   varbinary(127),
  `server_host_name`    varchar(255),
  `user`                varchar(127),
  `operation`           enum('UPLOAD','DOWNLOAD','UNKNOWN') NOT NULL,
  `cipher`              enum('NONE','AES128','AES256','UNKNOWN') NOT NULL,
  `source_paths`        text,
  `dest_paths`          text,
  `cookie`              text,
  `tags`                text,
  `token`               text,
  `server_docroot`      text,
  `client_docroot`      text,

  PRIMARY KEY (`id`),
  UNIQUE KEY `idx_transfer_id` (`transfer_id`),
  UNIQUE KEY `idx_stats_id` (`stats_id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

CREATE TABLE `fasp_stats` (
  `id`                  bigint          NOT NULL auto_increment,

  `seq_created`         bigint,
  `seq_modified`        bigint,

  `status`              enum('RUNNING','PAUSED','COMPLETED','CANCELLED','ERROR','WILLRETRY','ORPHANED','UNKNOWN') NOT NULL,
  `started_at`          datetime        NOT NULL,
  `stopped_at`          datetime,
  `updated_at`          datetime,

  `transport`           enum('FASP2','HTTP','UNKNOWN'),         /* transfer agg: last session */
  `files_complete`      int             NOT NULL DEFAULT 0,
  `files_failed`        int             NOT NULL DEFAULT 0,
  `files_attempted`     int             NOT NULL DEFAULT 0,
  `files_skipped`       int             NOT NULL DEFAULT 0,
  `bytes_written`       bigint          NOT NULL DEFAULT 0,
  `bytes_transferred`   bigint          NOT NULL DEFAULT 0,
  `bytes_lost`          bigint          NOT NULL DEFAULT 0,
  `elapsed_usec`        bigint          NOT NULL DEFAULT 0,
  `err_code`            int             NOT NULL,
  `err_desc`            varchar(255),
  `bytes_pretransfer`   bigint          NOT NULL DEFAULT 0,
  `files_pretransfer`   int             NOT NULL DEFAULT 0,
  `dirs_pretransfer`    int             NOT NULL DEFAULT 0,

  /*
   * Below values are actually "instantaneous" values that should rather be
   * stored as a series in a separate table. For now, we keep the "current"
   * value for ongoing transfers. Once the transfer is completed, these
   * columns will therefore contain the "last" value received.
   */
  `network_delay`       int             NOT NULL DEFAULT 0,
  `minimum_rate`        bigint          NOT NULL DEFAULT 0,     /* last rate reported */
  `target_rate`         bigint          NOT NULL DEFAULT 0,     /* last rate reported */
  `policy`              enum('FIXED','HIGH','FAIR','LOW','UNKNOWN') NOT NULL, /* last policy reported */

  PRIMARY KEY  (`id`),
  KEY `idx_seq_created` (`seq_created`),
  UNIQUE KEY `idx_seq_modified` (`seq_modified`),
  KEY `idx_status` (`status`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

CREATE TABLE `fasp_nodes` (
  `id`                  bigint          NOT NULL auto_increment,
  `created_at`          datetime        NOT NULL,
  `updated_at`          datetime        NOT NULL,
  `node_id`             varbinary(127)  NOT NULL,
  `cluster_id`          varbinary(127)  NOT NULL,
  `host_ip`             varbinary(127)  NOT NULL,
  `host_name`           varchar(255)    NOT NULL,
  `active_sessions`     int             DEFAULT 0,
  `node_status`         enum('ACTIVE','ERROR','DISABLED') DEFAULT 'ACTIVE',
  `err_desc`            varchar(255),                           /* null if node_status='ACTIVE' */
  `last_successful_poll` datetime,
  `reported_host_name`  varchar(255),
  `reported_os`         varchar(16),
  `reported_os_version` varchar(255),
  `reported_at`         datetime,

  PRIMARY KEY  (`id`),
  UNIQUE KEY `idx_host_ip` (`host_ip`),
  KEY `idx_host_name` (`host_name`),
  KEY `idx_cluster_id` (`cluster_id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;


CREATE TABLE `configuration` (
  `id`                  bigint          NOT NULL auto_increment,
  `name`                varchar(128)    NOT NULL,
  `value`               text            NOT NULL,
  PRIMARY KEY  (`id`),
  UNIQUE KEY `idx_name` (`name`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8; 


CREATE TABLE `api_commands` (
  `id`                  bigint          NOT NULL auto_increment,
  `command`             text            NOT NULL,
  `result`              text,

  `removal`             enum('ON_SUCCESS','ON_COMPLETION','NEVER') DEFAULT 'ON_COMPLETION',
  `status`              enum('PENDING','LOCKED','COMPLETED','FAILED') DEFAULT 'PENDING',

  `created_at`          datetime,
  `updated_at`          datetime,
  `locked_by`           varchar(127),

  PRIMARY KEY  (`id`),
  KEY `idx_status` (`status`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;

DROP TRIGGER IF EXISTS trig_api_commands_created;
DROP TRIGGER IF EXISTS trig_api_commands_modified;

DELIMITER ;;

CREATE TRIGGER trig_api_commands_created
BEFORE INSERT ON api_commands
FOR EACH ROW BEGIN
 SET NEW.created_at = NOW();
 SET NEW.updated_at = NOW();
END;;

CREATE TRIGGER trig_api_commands_modified
BEFORE UPDATE ON api_commands
FOR EACH ROW BEGIN
 SET NEW.updated_at = NOW();
END;;

DELIMITER ;


/********************** Internal stats-collector tables **********************/

CREATE TABLE sc_iterator_sequence (
  next bigint
) SELECT 1 AS next;


CREATE TABLE `sc_nodes_poll_status` (
  `id`                  bigint          NOT NULL,               /* foreign key to fasp_nodes.id */
  `poll_status`         enum('IDLE','LOCKED') DEFAULT 'IDLE',
  `locked_by`           varbinary(127),                         /* stats-collector ID */
  `iteration_token`     varbinary(255),
  `last_poll_ts`        bigint          DEFAULT 0,
  `keep_alive_ts`       bigint          DEFAULT 0,

  PRIMARY KEY  (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;


CREATE TABLE `sc_nodes_poll_config` (
  `id`                  bigint          NOT NULL auto_increment,
  `ref_uuid`            varbinary(127)  NOT NULL,
  `host_name`           varchar(255)    NOT NULL,
  `user_name`           varchar(255),
  `encrypted_password`  varchar(255),
  `port`                int             NOT NULL DEFAULT 0,
  `use_https`           tinyint         NOT NULL DEFAULT 0,
  `created_at`          datetime        NOT NULL,
  `updated_at`          datetime        NOT NULL,

  PRIMARY KEY (`id`),
  UNIQUE KEY `idx_ref_uuid` (`ref_uuid`),
  KEY `idx_host_name` (`host_name`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8;


CREATE TABLE `sc_version` (
  `schema_version`      bigint          NOT NULL,
  `created_at`          datetime        NOT NULL,
  `updated_at`          datetime        NOT NULL,
  `app_version`         varchar(255)    NOT NULL
) ENGINE=InnoDB DEFAULT CHARSET=utf8; 

SET TIME_ZONE=@OLD_TIME_ZONE;
SET SQL_MODE=@OLD_SQL_MODE;
SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS;
SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS;
SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT;
SET CHARACTER_SET_RESULTS=@OLD_CHARACTER_SET_RESULTS;
SET COLLATION_CONNECTION=@OLD_COLLATION_CONNECTION;
SET SQL_NOTES=@OLD_SQL_NOTES;
