#!/bin/bash

# Generic cross-platform daemon script functions.
# (c) Aspera Inc.

# -----------------------------------------------------------------------------
# Global variables that need to be set by the daemon script before calling the
# cdf_go function.
# -----------------------------------------------------------------------------
# CDF_DAEMON_NAME= (the name of the daemon - service XYZ start)
# CDF_DAEMON_DESC= (short description of what the daemon does)
# CDF_DAEMON_USER= (the user with which to launch the daemon)
# CDF_PID_FILE= (full path to PID file)
# CDF_EXECUTABLE= (the full path to the daemon executable)
# CDF_ARGUMENTS= (any arguments passed to the daemon executable on start-up)

# -----------------------------------------------------------------------------
# Options and switches.
# -----------------------------------------------------------------------------
CDF_NATIVE_STOP_ENABLED=false
CDF_THREAD_DUMP_ON_STOP=false

# -----------------------------------------------------------------------------
# Functions to be implemented (overridden) by the daemon script.
# -----------------------------------------------------------------------------

##
## Initializes $CDF_DAEMON_USER with the name of the user under which the daemon
## will be started.
##
cdf_init_user() { :; }

##
## Initializes $CDF_EXECUTABLE and $CDF_ARGUMENTS before starting the daemon.
## $CDF_EXECUTABLE is the full path to the daemon executable
## $CDF_ARGUMENTS contains the arguments passed to the daemon executable
##
cdf_init_start() { :; }

##
## Called for any unkown base command (other than start|stop|status|restart|threaddump)
##
cdf_handle_command() {
  cdf_usage "$@"
}


# -----------------------------------------------------------------------------
# Global variables
# -----------------------------------------------------------------------------

CDF_RETVAL=0
CDF_INSTALL_DIR=
CDF_PID=
CDF_EXECUTABLE=
CDF_ARGUMENTS=
CDF_SCRIPT=
CDF_ALL_ARGS="$0 $@"
CDF_USAGE="start [--no-daemon]|stop|status|restart|threaddump"
CDF_NO_DAEMON=false

# enable core dumps, limited to 5000 bytes
DAEMON_COREFILE_LIMIT=5000

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

##
## Initializes $CDF_INSTALL_DIR with the resolved directory path of the script's
## location. Symbolic links are resolved recursively.
##
cdf_init() {
  CDF_SCRIPT="$(readlink -f "$0")"
  CDF_INSTALL_DIR="$(dirname "${CDF_SCRIPT}")"
}

##
## Exits the script with the given exit code after printing an optional error
## message.
##
## $1: the exit code
## $2: an optional error message to print
##
cdf_error() {
  [ "$2" ] && echo "$2"
  cdf_log_invocation " ERROR" $1
  exit $1
}

##
## Find the pid of the program and signal it with SIGQUIT to produce a thread
## dump.
##
cdf_thread_dump() {
  cdf_daemon_status
  case "$CDF_RETVAL" in
    0|6)  echo "Generating thread dump..."
          cdf_thread_dump_of $CDF_PID
          sleep 1
          echo -n "Thread dump generated."
          [ -n "$STDOUT_FILE" ] && echo " See $STDOUT_FILE." || echo
          ;;
    *)    echo "$CDF_STATUS_MSG";;
  esac
}

##
## Signal the given pid with SIGQUIT to produce a Java thread dump.
##
## $1: pid of Java process
##
cdf_thread_dump_of() {
  kill -SIGQUIT $1
}

##
## Creates a directory (if it does not exist) and changes the ownership of the
## directory and all of its content recursively.
## Exists with an error if the directory cannot be created or it's ownership
## cannot be changed.
##
## $1: the directory to create / verify that it exists
## $2: (optional) the owner
## $3: (optional) the exit code, defaults to 64
##
cdf_create_dir() {
  local errcode=64
  if [ -n "$3" ]; then
    errcode=$3
  fi
  if [ ! -d "$1" ]; then
    mkdir -p "$1" >/dev/null 2>&1 || cdf_error $errcode "Cannot create directory $1"
  fi
  if [ -n "$2" ]; then
    chown "$2" -R "$1" >/dev/null 2>&1 || cdf_error $errcode "Cannot change ownership of $1 to $2"
  fi
}

##
## Checks whether the daemon is currently running by reading the process ID from
## the file $CDF_PID_FILE and ensuring that there is a running process with
## that process ID.
## If the daemon is running, the variable $CDF_PID is set to the process ID of
## the daemon, otherwise $CDF_PID is cleared.
##
## return code: 0 if running, 1 otherwise
##
cdf_check_running() {
  local cmd
  if [ -f $CDF_PID_FILE ]; then
    CDF_PID=$(cat $CDF_PID_FILE)
    cmd=$(ps -p "$CDF_PID" --no-heading)
    if [ -n "$cmd" ]; then
      return 0
    fi
  fi
  CDF_PID=
  return 1
}

##
## Initializes the daemon user and adapts file permissions on the installation
## directory accordingly.
##
cdf_check_user() {
  cdf_init_user

  if [ -n "$CDF_DAEMON_USER" ]; then
    # check that the user exists
    id -u $CDF_DAEMON_USER >/dev/null 2>&1 || cdf_error 64 "$CDF_DAEMON_USER is not a valid username."

    # Check installation folder ownership and set privileges if needed
    chown $CDF_DAEMON_USER -R $CDF_INSTALL_DIR >/dev/null 2>&1 || cdf_error 64 "Cannot change ownership of $CDF_INSTALL_DIR to $CDF_DAEMON_USER"
    chmod a+rx $CDF_INSTALL_DIR >/dev/null 2>&1 || cdf_error 64 "Cannot change rights on $CDF_INSTALL_DIR"
    chmod u+rwx $CDF_INSTALL_DIR >/dev/null 2>&1 || cdf_error 64 "Cannot change rights on $CDF_INSTALL_DIR"
  fi
}

##
## Prints usage information.
##
cdf_usage() {
  cdf_error 1 "$0 $CDF_USAGE"
}


# -----------------------------------------------------------------------------
# Default Implementations
# -----------------------------------------------------------------------------

##
## Parses additional start arguments
##
cdf_parse_start_args() {
  shift
  while (($#)); do
    case "$1" in
       --no-daemon)
          CDF_NO_DAEMON=true
          ;;
       *)
          cdf_usage "$@"
          ;;
    esac
    shift
  done
}

##
## Prepares starting the daemon:
## - retrieves and checks the user
## - initializes the daemon executable and parameters
##
cdf_prepare_start() {
  cdf_check_user
  cdf_init_start
  if $CDF_NO_DAEMON ; then
    CDF_ARGUMENTS="-nodetach $CDF_ARGUMENTS"
  fi
  cdf_daemon_status "fix_pid_file"
}

##
## Prepares stopping the daemon.
##
cdf_prepare_stop() {
  cdf_daemon_status "fix_pid_file"
  case "$CDF_RETVAL" in
    0|6)  [ "$CDF_THREAD_DUMP_ON_STOP" = "true" ] && cdf_thread_dump_of $CDF_PID
          ;;
  esac
}


##
## Ensures that no daemon processes are running anymore, killing them if
## necessary.
##
## Sets CDF_RETVAL to
## - 0 if no more daemon processes are running at the end of the call
## - != 0 otherwise (some processes could not be terminated)
##
cdf_ensure_stopped() {
  local out ret
  out=$(pgrep -lfx "-d, " "$CDF_DAEMON_NAME")
  ret=$?
  if [ "$ret" = 0 ]; then
    # at least one daemon process still running...
    echo "Failed to stop daemon normally. Remaining processes will be killed [$out]"
    pkill -KILL -fx "$CDF_DAEMON_NAME"
    CDF_RETVAL=$?
  else
    # no daemon processes running
    CDF_RETVAL=0
  fi
}


##
## Restarts the daemon.
##
cdf_restart() {
  cdf_stop
  # sleep 1
  cdf_log_invocation " STOP COMPLETED - CALLING START"
  cdf_start
}


##
## Determines the status of the deamon and
## * sets CDF_PID to the pid of the running daemon or clears it if the
##   daemon is not running.
## * sets CDF_RETVAL according to table below
## * sets CDF_STATUS_MSG according to table below
##
##   0 daemon running
##   1 daemon not running, pid file exists
##   2 unable to determine status
##   3 daemon not running, pid file does not exist
##   4 daemon not running, pid file does not exist or cannot be read
##   5 daemon not running, pid file points to another running process
##   6 daemon running, pid file points to another running process
##
## $1: if set to "fix_pid_file", will fix the incorrect pid file in case 5 & 6
##
cdf_daemon_status() {
  local pids_from_exe ret

  CDF_PID=
  CDF_RETVAL=
  CDF_STATUS_MSG=

  CDF_PID=$(pidofproc -p "$CDF_PID_FILE" "$CDF_EXECUTABLE" 2> /dev/null)
  CDF_RETVAL=$?

  case "$CDF_RETVAL" in
    0)  CDF_STATUS_MSG="$CDF_DAEMON_NAME running [pid $CDF_PID]";;
    1)  CDF_STATUS_MSG="$CDF_DAEMON_NAME not running, pid file exists";;
    3)  CDF_STATUS_MSG="$CDF_DAEMON_NAME not running, pid file does not exist";;
    4)  CDF_STATUS_MSG="$CDF_DAEMON_NAME not running, pid file does not exist or cannot be read";;
    *)  CDF_STATUS_MSG="$CDF_DAEMON_NAME not running ($CDF_RETVAL)";;
  esac

  pids_from_exe=$(pidofproc "$CDF_EXECUTABLE")
  ret=$?

  # echo "cdf_daemon_status 1: pids_from_exe=$pids_from_exe ret=$ret CDF_RETVAL=$CDF_RETVAL CDF_PID=$CDF_PID CDF_STATUS_MSG=$CDF_STATUS_MSG"

  if [ "$ret" != 0 ]; then
    # no daemon processes found running
    if [ "$CDF_RETVAL" = 0 ]; then
      CDF_RETVAL=5
      CDF_STATUS_MSG="$CDF_DAEMON_NAME not running, pid file points to another running process"
    fi
    # SUSE returns 1 instead of 0 if pid file points to another running process
    # To make sure we can still start the daemon (even though the status is not
    # reported 100% correctly), let's remove the file in any case if it exists...
    [ "$1" = "fix_pid_file" ] && rm -f "$CDF_PID_FILE" 2> /dev/null
    return
  fi

  # one or several daemon processes are running
  for i in $pids_from_exe; do
    [ "$CDF_PID" = "$i" ] && return ## the pid belongs to "our" daemon
  done

  cdf_find_child_process "$pids_from_exe"
  # write it into pid file if desired
  if [ "$1" = "fix_pid_file" ]; then
    cdf_log "fixing pidfile with $CDF_PID"
    echo "$CDF_PID" > "$CDF_PID_FILE"
  fi

  if [ "$CDF_RETVAL" = 0 ]; then
    CDF_STATUS_MSG="$CDF_DAEMON_NAME running, pid file points to another running process"
  else
    CDF_STATUS_MSG="$CDF_DAEMON_NAME running, pid file does not exist or cannot be read"
  fi
  CDF_RETVAL=6
}


##
## Finds the process in a list of processes that is a child of another process
## in the list. If no child is found, the last pid in the list is returned.
##
## $1: space separated list of pids, e.g. "2653 35643"
##
cdf_find_child_process() {
  local ppid
  for i in $1; do
    CDF_PID=$i
    ppid=$(ps -p $i -o ppid=)
    for j in $1; do
      [ $j = $ppid ] && return
    done
  done
}


# -----------------------------------------------------------------------------
# Main logic
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# OS specific overrides
# -----------------------------------------------------------------------------

if test `uname` = "Darwin"; then
  # ---------------------------------------------------------------------------
  # Mac OS X
  # ---------------------------------------------------------------------------
  echo "Mac OS X not yet supported"

elif [ -f /etc/SuSE-release -o -f /etc/SUSE-brand ] ; then
  # ---------------------------------------------------------------------------
  # SuSE Linux
  # ---------------------------------------------------------------------------

  . /etc/rc.status

  rc_reset

  cdf_start() {
    echo -n $"Starting $CDF_DAEMON_NAME: "
    cdf_prepare_start
    case "$CDF_RETVAL" in
      0|6)  echo -n "already running [pid $CDF_PID]"
            CDF_RETVAL=0
            cdf_report_status
            ;;
      *)    local args="-p "$CDF_PID_FILE" "$CDF_EXECUTABLE""
            [ -n "$CDF_DAEMON_USER" ] && args="-u $CDF_DAEMON_USER $args"
			if [ -n "$CDF_DAEMON_USER" ]; then
				$CDF_EXECUTABLE $CDF_ARGUMENTS
			else
				$CDF_EXECUTABLE $CDF_ARGUMENTS
			fi
            rc_status -v
            ;;
    esac
  }

  cdf_stop() {
    echo -n $"Stopping: $CDF_DAEMON_NAME: "
    cdf_prepare_stop

    if $CDF_NATIVE_STOP_ENABLED ; then
      echo -n $"Stopping: $CDF_DAEMON_NAME: "
      killproc -SIGINT -p "$CDF_PID_FILE" "$CDF_EXECUTABLE"
      rc_status -v
    else
      case "$CDF_RETVAL" in
        0|6)
            [ "$CDF_RETVAL" == 6 ] && echo -n "$CDF_STATUS_MSG"
            "$CDF_EXECUTABLE" -stop -pidfile "$CDF_PID_FILE" "$DAEMON_MAIN_CLASS"
            cdf_ensure_stopped
            ;;
        *)  echo -n "$CDF_STATUS_MSG"
            CDF_RETVAL=0
            ;;
      esac
      cdf_report_status
    fi
  }

  cdf_status() {
    cdf_daemon_status
    echo -n "$CDF_STATUS_MSG"
    cdf_report_status
  }

  cdf_report_status() {
    cdf_map_status
    rc_status -v
  }

  cdf_map_status() {
    case "$CDF_RETVAL" in
      6)  CDF_RETVAL=0;;
    esac
    return $CDF_RETVAL
  }

elif [[ -f "/etc/redhat-release"\
     || -f "/etc/yellowdog-release"\
     || -f "/etc/mandrake-release"\
     || -f "/etc/system-release" && $(cat /etc/system-release | awk '{print $1 $2}') = "AmazonLinux" ]] ; then
  # ---------------------------------------------------------------------------
  # RedHat, Fedora, CentOS, Amazon
  # ---------------------------------------------------------------------------

  # Source function library
  . /etc/rc.d/init.d/functions

  cdf_start() {
    echo -n $"Starting $CDF_DAEMON_NAME: "
    cdf_prepare_start
    case "$CDF_RETVAL" in
      0|6)  echo -n "already running [pid $CDF_PID]"
            warning
            echo
            CDF_RETVAL=0
            return
            ;;
    esac

    if [ -n "$CDF_DAEMON_USER" ]; then
       $CDF_EXECUTABLE $CDF_ARGUMENTS
    else
       $CDF_EXECUTABLE $CDF_ARGUMENTS
    fi
    CDF_RETVAL=$?
    cdf_daemon_status
    case "$CDF_RETVAL" in
      0|6)  success
            CDF_RETVAL=0;;
      *)    failure;;
    esac
    echo
  }

  cdf_stop() {
    echo -n $"Stopping $CDF_DAEMON_NAME: "
    cdf_prepare_stop

    if $CDF_NATIVE_STOP_ENABLED ; then
      killproc -p "$CDF_PID_FILE" "$CDF_EXECUTABLE"
      CDF_RETVAL=$?
      echo
    else
      case "$CDF_RETVAL" in
        0|6)
            [ "$CDF_RETVAL" == 6 ] && echo && echo -n "$CDF_STATUS_MSG"
            "$CDF_EXECUTABLE" -stop -pidfile "$CDF_PID_FILE" "$DAEMON_MAIN_CLASS"
            cdf_ensure_stopped
            [ "$CDF_RETVAL" = 0 ] && success || failure
            ;;
        *)  echo
            echo -n "$CDF_STATUS_MSG"
            warning
            CDF_RETVAL=0
            ;;
      esac
      echo
    fi
  }

  cdf_status() {
    cdf_daemon_status
    echo -n "$CDF_STATUS_MSG"
    case "$CDF_RETVAL" in
      0|6)  success
            CDF_RETVAL=0;;
      *)    warning;;
    esac
    echo
  }

else
  # ---------------------------------------------------------------------------
  # All other distributions
  # ---------------------------------------------------------------------------

  # load logging functions
  . /lib/lsb/init-functions


  cdf_start() {
    log_daemon_msg "Starting $CDF_DAEMON_DESC: $CDF_DAEMON_NAME"
    cdf_prepare_start

    local args="--start --exec "$CDF_EXECUTABLE" --oknodo"
    [ -n "$CDF_DAEMON_USER" ] && args="-c $CDF_DAEMON_USER $args"
    if [ -n "$CDF_DAEMON_USER" ]; then
       $CDF_EXECUTABLE $CDF_ARGUMENTS
    else
       $CDF_EXECUTABLE $CDF_ARGUMENTS
    fi
    CDF_RETVAL=$?
    log_end_msg $CDF_RETVAL
  }

  cdf_stop() {
    log_daemon_msg "Stopping $CDF_DAEMON_DESC: $CDF_DAEMON_NAME"
    cdf_prepare_stop
    if $CDF_NATIVE_STOP_ENABLED ; then
      start-stop-daemon --stop --exec "$CDF_EXECUTABLE" --retry TERM/5/TERM/20/KILL/5 --oknodo
      CDF_RETVAL=$?
    else
      case "$CDF_RETVAL" in
        0|6)
            [ "$CDF_RETVAL" == 6 ] && log_daemon_msg "$CDF_STATUS_MSG"
            "$CDF_EXECUTABLE" -stop -pidfile "$CDF_PID_FILE" "$DAEMON_MAIN_CLASS"
            cdf_ensure_stopped
            ;;
        *)  log_failure_msg "$CDF_STATUS_MSG"
            CDF_RETVAL=0
            ;;
      esac
    fi
    log_end_msg $CDF_RETVAL
  }

  cdf_status() {
    cdf_daemon_status
    case "$CDF_RETVAL" in
      0|6)  log_success_msg "$CDF_STATUS_MSG"
            CDF_RETVAL=0;;
      *)    log_failure_msg "$CDF_STATUS_MSG";;
    esac
  }

fi


cdf_log_invocation() {
  local PS_MSG=
  local PID_MSG=
  local EXIT_MSG=""

  [ -n "$2" ] && EXIT_MSG=" exit status=$2"
  cdf_create_dir "$LOG_DIR" "$DAEMON_USER"

  PS_MSG="$(ps -FC "$CDF_DAEMON_NAME")"
  PID_MSG="PID from $CDF_PID_FILE: [$(cat "$CDF_PID_FILE" 2>/dev/null)]"

  cdf_log "\n########## $(date --rfc-3339=ns) COMMAND [$CDF_ALL_ARGS]$1$EXIT_MSG\n$PS_MSG\n$PID_MSG\n##########"
}

cdf_log() {
  echo -e "$1" >> "$STDOUT_FILE"
}

cdf_go() {
  # Exit if the package has been removed or not correctly installed.
  [ -x "$CDF_EXECUTABLE" ] || cdf_error 1 "The daemon $CDF_EXECUTABLE does not exist or is not executable"

  case "$1" in
     start)
        cdf_log_invocation
        cdf_parse_start_args "$@"
        cdf_start
        cdf_log_invocation " COMPLETED" $CDF_RETVAL
        ;;
     stop)
        cdf_log_invocation
        cdf_stop
        cdf_log_invocation " COMPLETED" $CDF_RETVAL
        ;;
     status)
        cdf_status
        ;;
     restart)
        cdf_log_invocation
        cdf_restart
        cdf_log_invocation " COMPLETED" $CDF_RETVAL
        ;;
     threaddump)
        cdf_log_invocation
        cdf_thread_dump
        cdf_log_invocation " COMPLETED" $CDF_RETVAL
        ;;
     *)
        cdf_log_invocation
        cdf_handle_command "$@"
        cdf_log_invocation " COMPLETED" $CDF_RETVAL
        ;;
  esac

  exit $CDF_RETVAL
}
